@@ -129,6 +129,11 @@
 /*	Optional lookup tables that map names received from remote SMTP
 /*	clients via the TLS Server Name Indication (SNI) extension to the
 /*	appropriate keys and certificate chains.
+/* .PP
+/*	Available in Postfix 3.5, 3.4.6, 3.3.5, 3.2.10, 3.1.13 and later:
+/* .IP "\fBtls_fast_shutdown_enable (yes)\fR"
+/*	A workaround for implementations that hang Postfix while shutting
+/*	down a TLS session, until Postfix times out.
 /* STARTTLS SERVER CONTROLS
 /* .ad
 /* .fi
@@ -613,11 +618,11 @@ static int tlsp_eval_tls_error(TLSP_STATE *state, int err)
     switch (err) {
 
 	/*
-	 * No error from SSL_read and SSL_write means that the plaintext
-	 * output buffer is full and that the plaintext input buffer is
-	 * empty. Stop read/write events on the ciphertext stream. Keep the
-	 * timer alive as a safety mechanism for the case that the plaintext
-	 * pseudothreads get stuck.
+	 * No error means a successful SSL_accept/connect/shutdown request or
+	 * sequence of SSL_read/write requests. Disable read/write events on
+	 * the ciphertext stream. Keep the ciphertext stream timer alive as a
+	 * safety mechanism for the case that the plaintext pseudothreads get
+	 * stuck.
 	 */
     case SSL_ERROR_NONE:
 	if (state->ssl_last_err != SSL_ERROR_NONE) {
@@ -671,10 +676,23 @@ static int tlsp_eval_tls_error(TLSP_STATE *state, int err)
     default:
 
 	/*
-	 * Allow buffered-up plaintext output to trickle out.
+	 * Allow buffered-up plaintext output to trickle out. Permanently
+	 * disable read/write activity on the ciphertext stream, so that this
+	 * function will no longer be called. Keep the ciphertext stream
+	 * timer alive as a safety mechanism for the case that the plaintext
+	 * pseudothreads get stuck. Return into tlsp_strategy(), which will
+	 * enable plaintext write events.
 	 */
-	if (state->plaintext_buf && NBBIO_WRITE_PEND(state->plaintext_buf))
+#define TLSP_CAN_TRICKLE_OUT_PLAINTEXT(buf) \
+	((buf) && !NBBIO_ERROR_FLAGS(buf) && NBBIO_WRITE_PEND(buf))
+
+	if (TLSP_CAN_TRICKLE_OUT_PLAINTEXT(state->plaintext_buf)) {
+	    event_disable_readwrite(ciphertext_fd);
+	    event_request_timer(tlsp_ciphertext_event, (void *) state,
+				state->timeout);
+	    state->flags |= TLSP_FLAG_NO_MORE_CIPHERTEXT_IO;
 	    return (TLSP_STAT_OK);
+	}
 	tlsp_state_free(state);
 	return (TLSP_STAT_ERR);
     }
@@ -745,119 +763,153 @@ static void tlsp_strategy(TLSP_STATE *state)
     int     handshake_err;
 
     /*
-     * Do not enable plain-text I/O before completing the TLS handshake.
-     * Otherwise the remote peer can prepend plaintext to the optional
-     * TLS_SESS_STATE object.
+     * This function is called after every ciphertext or plaintext event, to
+     * schedule new ciphertext or plaintext I/O.
      */
-    if (state->flags & TLSP_FLAG_DO_HANDSHAKE) {
-	state->timeout = state->handshake_timeout;
-	if (state->is_server_role)
-	    ssl_stat = SSL_accept(tls_context->con);
-	else
-	    ssl_stat = SSL_connect(tls_context->con);
-	if (ssl_stat != 1) {
-	    handshake_err = SSL_get_error(tls_context->con, ssl_stat);
-	    tlsp_eval_tls_error(state, handshake_err);
-	    /* At this point, state could be a dangling pointer. */
+
+    /*
+     * Try to make an SSL I/O request. If this fails with SSL_ERROR_WANT_READ
+     * or SSL_ERROR_WANT_WRITE, enable ciphertext read or write events, and
+     * retry the SSL I/O request in a later tlsp_strategy() call.
+     */
+    if ((state->flags & TLSP_FLAG_NO_MORE_CIPHERTEXT_IO) == 0) {
+
+	/*
+	 * Do not enable plain-text I/O before completing the TLS handshake.
+	 * Otherwise the remote peer can prepend plaintext to the optional
+	 * TLS_SESS_STATE object.
+	 */
+	if (state->flags & TLSP_FLAG_DO_HANDSHAKE) {
+	    state->timeout = state->handshake_timeout;
+	    ERR_clear_error();
+	    if (state->is_server_role)
+		ssl_stat = SSL_accept(tls_context->con);
+	    else
+		ssl_stat = SSL_connect(tls_context->con);
+	    if (ssl_stat != 1) {
+		handshake_err = SSL_get_error(tls_context->con, ssl_stat);
+		tlsp_eval_tls_error(state, handshake_err);
+		/* At this point, state could be a dangling pointer. */
+		return;
+	    }
+	    state->flags &= ~TLSP_FLAG_DO_HANDSHAKE;
+	    state->timeout = state->session_timeout;
+	    if (tlsp_post_handshake(state) != TLSP_STAT_OK) {
+		/* At this point, state is a dangling pointer. */
+		return;
+	    }
+	}
+
+	/*
+	 * Shutdown and self-destruct after NBBIO error. This automagically
+	 * cleans up all pending read/write and timeout event requests.
+	 * Before shutting down TLS, we stop all plain-text I/O events but
+	 * keep the NBBIO error flags.
+	 */
+	plaintext_buf = state->plaintext_buf;
+	if (NBBIO_ERROR_FLAGS(plaintext_buf)) {
+	    if (NBBIO_ACTIVE_FLAGS(plaintext_buf))
+		nbbio_disable_readwrite(state->plaintext_buf);
+	    ERR_clear_error();
+	    if (!SSL_in_init(tls_context->con)
+		&& (ssl_stat = SSL_shutdown(tls_context->con)) < 0) {
+		handshake_err = SSL_get_error(tls_context->con, ssl_stat);
+		tlsp_eval_tls_error(state, handshake_err);
+		/* At this point, state could be a dangling pointer. */
+		return;
+	    }
+	    tlsp_state_free(state);
 	    return;
 	}
-	state->flags &= ~TLSP_FLAG_DO_HANDSHAKE;
-	state->timeout = state->session_timeout;
-	if (tlsp_post_handshake(state) != TLSP_STAT_OK) {
+
+	/*
+	 * Try to move data from the plaintext input buffer to the TLS
+	 * engine.
+	 * 
+	 * XXX We're supposed to repeat the exact same SSL_write() call
+	 * arguments after an SSL_ERROR_WANT_READ or SSL_ERROR_WANT_WRITE
+	 * result. Rumor has it that this is because each SSL_write() call
+	 * reads from the buffer incrementally, and returns > 0 only after
+	 * the final byte is processed. Rumor also has it that setting
+	 * SSL_MODE_ENABLE_PARTIAL_WRITE and
+	 * SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER voids this requirement, and
+	 * that repeating the request with an increased request size is OK.
+	 * Unfortunately all this is not or poorly documented, and one has to
+	 * rely on statements from OpenSSL developers in public mailing
+	 * archives.
+	 */
+	ssl_write_err = SSL_ERROR_NONE;
+	while (NBBIO_READ_PEND(plaintext_buf) > 0) {
+	    ERR_clear_error();
+	    ssl_stat = SSL_write(tls_context->con, NBBIO_READ_BUF(plaintext_buf),
+				 NBBIO_READ_PEND(plaintext_buf));
+	    ssl_write_err = SSL_get_error(tls_context->con, ssl_stat);
+	    if (ssl_write_err != SSL_ERROR_NONE)
+		break;
+	    /* Allow the plaintext pseudothread to read more data. */
+	    NBBIO_READ_PEND(plaintext_buf) -= ssl_stat;
+	    if (NBBIO_READ_PEND(plaintext_buf) > 0)
+		memmove(NBBIO_READ_BUF(plaintext_buf),
+			NBBIO_READ_BUF(plaintext_buf) + ssl_stat,
+			NBBIO_READ_PEND(plaintext_buf));
+	}
+
+	/*
+	 * Try to move data from the TLS engine to the plaintext output
+	 * buffer. Note: data may arrive as a side effect of calling
+	 * SSL_write(), therefore we call SSL_read() after calling
+	 * SSL_write().
+	 * 
+	 * XXX We're supposed to repeat the exact same SSL_read() call arguments
+	 * after an SSL_ERROR_WANT_READ or SSL_ERROR_WANT_WRITE result. This
+	 * supposedly means that our plaintext writer must not memmove() the
+	 * plaintext output buffer until after the SSL_read() call succeeds.
+	 * For now I'll ignore this, because 1) SSL_read() is documented to
+	 * return the bytes available, instead of returning > 0 only after
+	 * the entire buffer is processed like SSL_write() does; and 2) there
+	 * is no "read" equivalent of the SSL_R_BAD_WRITE_RETRY,
+	 * SSL_MODE_ENABLE_PARTIAL_WRITE or
+	 * SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER features.
+	 */
+	ssl_read_err = SSL_ERROR_NONE;
+	while (NBBIO_WRITE_PEND(state->plaintext_buf) < NBBIO_BUFSIZE(plaintext_buf)) {
+	    ERR_clear_error();
+	    ssl_stat = SSL_read(tls_context->con,
+				NBBIO_WRITE_BUF(plaintext_buf)
+				+ NBBIO_WRITE_PEND(state->plaintext_buf),
+				NBBIO_BUFSIZE(plaintext_buf)
+				- NBBIO_WRITE_PEND(state->plaintext_buf));
+	    ssl_read_err = SSL_get_error(tls_context->con, ssl_stat);
+	    if (ssl_read_err != SSL_ERROR_NONE)
+		break;
+	    NBBIO_WRITE_PEND(plaintext_buf) += ssl_stat;
+	}
+
+	/*
+	 * Try to enable/disable ciphertext read/write events. If SSL_write()
+	 * was satisfied, see if SSL_read() wants to do some work. In case of
+	 * an unrecoverable error, this automagically destroys the session
+	 * state after cleaning up all pending read/write and timeout event
+	 * requests.
+	 */
+	if (tlsp_eval_tls_error(state, ssl_write_err != SSL_ERROR_NONE ?
+				ssl_write_err : ssl_read_err) < 0)
 	    /* At this point, state is a dangling pointer. */
 	    return;
-	}
     }
 
     /*
-     * Shutdown and self-destruct after NBBIO error. This automagically
-     * cleans up all pending read/write and timeout event requests. Before
-     * shutting down TLS, we stop all plain-text I/O events but keep the
-     * NBBIO error flags.
+     * Destroy state when the ciphertext I/O was permanently disabled and we
+     * can no longer trickle out plaintext.
      */
-    plaintext_buf = state->plaintext_buf;
-    if (NBBIO_ERROR_FLAGS(plaintext_buf)) {
-	if (NBBIO_ACTIVE_FLAGS(plaintext_buf))
-	    nbbio_disable_readwrite(state->plaintext_buf);
-	ssl_stat = SSL_shutdown(tls_context->con);
-	/* XXX Wait for return value 1 if sessions are to be reused? */
-	if (ssl_stat < 0) {
-	    handshake_err = SSL_get_error(tls_context->con, ssl_stat);
-	    tlsp_eval_tls_error(state, handshake_err);
-	    /* At this point, state could be a dangling pointer. */
+    else {
+	plaintext_buf = state->plaintext_buf;
+	if (!TLSP_CAN_TRICKLE_OUT_PLAINTEXT(plaintext_buf)) {
+	    tlsp_state_free(state);
 	    return;
 	}
-	tlsp_state_free(state);
-	return;
     }
 
-    /*
-     * Try to move data from the plaintext input buffer to the TLS engine.
-     * 
-     * XXX We're supposed to repeat the exact same SSL_write() call arguments
-     * after an SSL_ERROR_WANT_READ or SSL_ERROR_WANT_WRITE result. Rumor has
-     * it that this is because each SSL_write() call reads from the buffer
-     * incrementally, and returns > 0 only after the final byte is processed.
-     * Rumor also has it that setting SSL_MODE_ENABLE_PARTIAL_WRITE and
-     * SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER voids this requirement, and that
-     * repeating the request with an increased request size is OK.
-     * Unfortunately all this is not or poorly documented, and one has to
-     * rely on statements from OpenSSL developers in public mailing archives.
-     */
-    ssl_write_err = SSL_ERROR_NONE;
-    while (NBBIO_READ_PEND(plaintext_buf) > 0) {
-	ssl_stat = SSL_write(tls_context->con, NBBIO_READ_BUF(plaintext_buf),
-			     NBBIO_READ_PEND(plaintext_buf));
-	ssl_write_err = SSL_get_error(tls_context->con, ssl_stat);
-	if (ssl_write_err != SSL_ERROR_NONE)
-	    break;
-	/* Allow the plaintext pseudothread to read more data. */
-	NBBIO_READ_PEND(plaintext_buf) -= ssl_stat;
-	if (NBBIO_READ_PEND(plaintext_buf) > 0)
-	    memmove(NBBIO_READ_BUF(plaintext_buf),
-		    NBBIO_READ_BUF(plaintext_buf) + ssl_stat,
-		    NBBIO_READ_PEND(plaintext_buf));
-    }
-
-    /*
-     * Try to move data from the TLS engine to the plaintext output buffer.
-     * Note: data may arrive as a side effect of calling SSL_write(),
-     * therefore we call SSL_read() after calling SSL_write().
-     * 
-     * XXX We're supposed to repeat the exact same SSL_read() call arguments
-     * after an SSL_ERROR_WANT_READ or SSL_ERROR_WANT_WRITE result. This
-     * supposedly means that our plaintext writer must not memmove() the
-     * plaintext output buffer until after the SSL_read() call succeeds. For
-     * now I'll ignore this, because 1) SSL_read() is documented to return
-     * the bytes available, instead of returning > 0 only after the entire
-     * buffer is processed like SSL_write() does; and 2) there is no "read"
-     * equivalent of the SSL_R_BAD_WRITE_RETRY, SSL_MODE_ENABLE_PARTIAL_WRITE
-     * or SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER features.
-     */
-    ssl_read_err = SSL_ERROR_NONE;
-    while (NBBIO_WRITE_PEND(state->plaintext_buf) < NBBIO_BUFSIZE(plaintext_buf)) {
-	ssl_stat = SSL_read(tls_context->con,
-			    NBBIO_WRITE_BUF(plaintext_buf)
-			    + NBBIO_WRITE_PEND(state->plaintext_buf),
-			    NBBIO_BUFSIZE(plaintext_buf)
-			    - NBBIO_WRITE_PEND(state->plaintext_buf));
-	ssl_read_err = SSL_get_error(tls_context->con, ssl_stat);
-	if (ssl_read_err != SSL_ERROR_NONE)
-	    break;
-	NBBIO_WRITE_PEND(plaintext_buf) += ssl_stat;
-    }
-
-    /*
-     * Try to enable/disable ciphertext read/write events. If SSL_write() was
-     * satisfied, see if SSL_read() wants to do some work. In case of an
-     * unrecoverable error, this automagically destroys the session state
-     * after cleaning up all pending read/write and timeout event requests.
-     */
-    if (tlsp_eval_tls_error(state, ssl_write_err != SSL_ERROR_NONE ?
-			    ssl_write_err : ssl_read_err) < 0)
-	/* At this point, state is a dangling pointer. */
-	return;
-
     /*
      * Try to enable/disable plaintext read/write events. Basically, if we
      * have nothing to write to the plaintext stream, see if there is
@@ -877,13 +929,11 @@ static void tlsp_strategy(TLSP_STATE *state)
     if (NBBIO_WRITE_PEND(plaintext_buf) > 0) {
 	if (NBBIO_ACTIVE_FLAGS(plaintext_buf) & NBBIO_FLAG_READ)
 	    nbbio_disable_readwrite(plaintext_buf);
-	if ((NBBIO_ACTIVE_FLAGS(plaintext_buf) & NBBIO_FLAG_WRITE) == 0)
-	    nbbio_enable_write(plaintext_buf, state->timeout);
+	nbbio_enable_write(plaintext_buf, state->timeout);
     } else if (NBBIO_READ_PEND(plaintext_buf) < NBBIO_BUFSIZE(plaintext_buf)) {
 	if (NBBIO_ACTIVE_FLAGS(plaintext_buf) & NBBIO_FLAG_WRITE)
 	    nbbio_disable_readwrite(plaintext_buf);
-	if ((NBBIO_ACTIVE_FLAGS(plaintext_buf) & NBBIO_FLAG_READ) == 0)
-	    nbbio_enable_read(plaintext_buf, state->timeout);
+	nbbio_enable_read(plaintext_buf, state->timeout);
     } else {
 	if (NBBIO_ACTIVE_FLAGS(plaintext_buf))
 	    nbbio_slumber(plaintext_buf, state->timeout);
@@ -946,13 +996,7 @@ static int tlsp_client_start_pre_handshake(TLSP_STATE *state)
 {
     state->client_start_props->ctx = state->appl_state;
     state->client_start_props->fd = state->ciphertext_fd;
-    /* These predicates and warning belong inside tls_client_start(). */
-    if (!TLS_DANE_BASED(state->client_start_props->tls_level)
-	|| tls_dane_avail())
-	state->tls_context = tls_client_start(state->client_start_props);
-    else
-	msg_warn("%s: DANE requested, but not available",
-		 state->client_start_props->namaddr);
+    state->tls_context = tls_client_start(state->client_start_props);
     if (state->tls_context != 0)
 	return (TLSP_STAT_OK);
 
@@ -1144,8 +1188,7 @@ static void tlsp_log_config_diff(const char *server_cfg, const char *client_cfg)
 /* tlsp_client_init - initialize a TLS client engine */
 
 static TLS_APPL_STATE *tlsp_client_init(TLS_CLIENT_PARAMS *tls_params,
-				          TLS_CLIENT_INIT_PROPS *init_props,
-					        int dane_based)
+				          TLS_CLIENT_INIT_PROPS *init_props)
 {
     TLS_APPL_STATE *appl_state;
     VSTRING *param_buf;
@@ -1178,8 +1221,8 @@ static TLS_APPL_STATE *tlsp_client_init(TLS_CLIENT_PARAMS *tls_params,
     init_key = tls_proxy_client_init_with_names_to_string(
 						      init_buf, init_props);
     init_buf_for_hashing = vstring_alloc(100);
-    init_key_for_hashing = STR(vstring_sprintf(init_buf_for_hashing, "%s%d\n",
-					       init_key, dane_based));
+    init_key_for_hashing = STR(vstring_sprintf(init_buf_for_hashing, "%s\n",
+					       init_key));
     if (tlsp_pre_jail_done == 0) {
 	if (tlsp_pre_jail_client_param_key == 0
 	    || tlsp_pre_jail_client_init_key == 0) {
@@ -1376,8 +1419,7 @@ static void tlsp_get_request_event(int event, void *context)
 	    return;
 	}
 	state->appl_state = tlsp_client_init(state->tls_params,
-					     state->client_init_props,
-		      TLS_DANE_BASED(state->client_start_props->tls_level));
+					     state->client_init_props);
 	ready = state->appl_state != 0;
 	break;
     case TLS_PROXY_FLAG_ROLE_SERVER:
@@ -1447,16 +1489,15 @@ static void tlsp_service(VSTREAM *plaintext_stream,
 			    TLSP_INIT_TIMEOUT, (void *) state);
 }
 
-/* pre_jail_init - pre-jail initialization */
+/* pre_jail_init_server - pre-jail initialization */
 
-static void pre_jail_init(char *unused_name, char **unused_argv)
+static void pre_jail_init_server(void)
 {
     TLS_SERVER_INIT_PROPS props;
     const char *cert_file;
     int     have_server_cert;
     int     no_server_cert_ok;
     int     require_server_cert;
-    int     clnt_use_tls;
 
     /*
      * The code in this routine is pasted literally from smtpd(8). I am not
@@ -1489,7 +1530,7 @@ static void pre_jail_init(char *unused_name, char **unused_argv)
     }
     var_tlsp_use_tls = var_tlsp_use_tls || var_tlsp_enforce_tls;
     if (!var_tlsp_use_tls) {
-	msg_warn("TLS service is requested, but disabled with %s or %s",
+	msg_warn("TLS server role is disabled with %s or %s",
 		 VAR_TLSP_TLS_LEVEL, VAR_TLSP_USE_TLS);
 	return;
     }
@@ -1580,6 +1621,13 @@ static void pre_jail_init(char *unused_name, char **unused_argv)
 	SSL_CTX_set_mode(tlsp_server_ctx->ssl_ctx,
 			 SSL_MODE_ENABLE_PARTIAL_WRITE
 			 | SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER);
+}
+
+/* pre_jail_init_client - pre-jail initialization */
+
+static void pre_jail_init_client(void)
+{
+    int     clnt_use_tls;
 
     /*
      * The cache with TLS_APPL_STATE instances for different TLS_CLIENT_INIT
@@ -1658,7 +1706,6 @@ static void pre_jail_init(char *unused_name, char **unused_argv)
     if (clnt_use_tls || var_tlsp_clnt_per_site[0] || var_tlsp_clnt_policy[0]) {
 	TLS_CLIENT_PARAMS tls_params;
 	TLS_CLIENT_INIT_PROPS init_props;
-	int     dane_based_mode;
 
 	tls_pre_jail_init(TLS_ROLE_CLIENT);
 
@@ -1685,12 +1732,21 @@ static void pre_jail_init(char *unused_name, char **unused_argv)
 				    CAfile = var_tlsp_clnt_CAfile,
 				    CApath = var_tlsp_clnt_CApath,
 				    mdalg = var_tlsp_clnt_fpt_dgst);
-	for (dane_based_mode = 0; dane_based_mode < 2; dane_based_mode++) {
-	    if (tlsp_client_init(&tls_params, &init_props,
-				 dane_based_mode) == 0)
-		msg_warn("TLS client initialization failed");
-	}
+	if (tlsp_client_init(&tls_params, &init_props) == 0)
+	    msg_warn("TLS client initialization failed");
     }
+}
+
+/* pre_jail_init - pre-jail initialization */
+
+static void pre_jail_init(char *unused_name, char **unused_argv)
+{
+
+    /*
+     * Initialize roles separately.
+     */
+    pre_jail_init_server();
+    pre_jail_init_client();
 
     /*
      * tlsp_client_init() needs to know if it is called pre-jail or
